from typing import Any, Dict, List, Tuple
import click

from . import config as peddler_config
from . import env
from . import exceptions
from . import fmt
from .__about__ import __version__


def update(root: str, interactive: bool = True) -> Dict[str, Any]:
    """
    Load and save the configuration.
    """
    config, defaults = load_all(root, interactive=interactive)
    peddler_config.save_config_file(root, config)
    peddler_config.merge(config, defaults)
    return config


def load_all(
    root: str, interactive: bool = True
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Load configuration and interactively ask questions to collect param values from the user.
    """
    config, defaults = peddler_config.load_all(root)
    if interactive:
        ask_questions(config, defaults)
    return config, defaults


def ask_questions(config: Dict[str, Any], defaults: Dict[str, Any]) -> None:
    run_for_prod = config.get("STORE_HOST") != "localhost:8000/"
    run_for_prod = click.confirm(
        fmt.question(
            "Are you configuring a production ready store? Type 'n' if you are just testing Peddler on your local computer"
        ),
        prompt_suffix=" ",
        default=run_for_prod,
    )
    if not run_for_prod:
        dev_values = {
            "STORE_HOST": "localhost:8000/",
            "ENABLE_HTTPS": False,
        }
        fmt.echo_info(
            """As you are not running this store in production, we automatically set the following configuration values:"""
        )
        for k, v in dev_values.items():
            config[k] = v
            fmt.echo_info("    {} = {}".format(k, v))

    if run_for_prod:
        ask("Your shopping cart domain name ", "STORE_HOST", config, defaults)
        if "localhost" in config["STORE_HOST"]:
            raise exceptions.PeddlerError(
                "You may not use 'localhost' as the store domain name. To run a local store for testing purposes you should answer 'n' to the previous question."
            )

    ask("Your store name/title", "PLATFORM_NAME", config, defaults)
    ask("Your new admin account username", "OPENCART_ADMIN_USERNAME", config, defaults)
    ask("Your new admin account password", "OPENCART_ADMIN_PASSWORD", config, defaults)
    ask("Email address for admin account", "CONTACT_EMAIL", config, defaults)

    if run_for_prod:
        ask_bool(
            (
                "Activate SSL/TLS certificates for HTTPS access? Important note:"
                " this will NOT work in a development environment."
            ),
            "ENABLE_HTTPS",
            config,
            defaults,
        )


def ask(
    question: str, key: str, config: Dict[str, Any], defaults: Dict[str, Any]
) -> None:
    default = env.render_str(config, config.get(key, defaults[key]))
    config[key] = click.prompt(
        fmt.question(question), prompt_suffix=" ", default=default, show_default=True
    )


def ask_bool(
    question: str, key: str, config: Dict[str, Any], defaults: Dict[str, Any]
) -> None:
    default = config.get(key, defaults[key])
    config[key] = click.confirm(
        fmt.question(question), prompt_suffix=" ", default=default
    )


def ask_choice(
    question: str,
    key: str,
    config: Dict[str, Any],
    defaults: Dict[str, Any],
    choices: List[str],
) -> None:
    default = config.get(key, defaults[key])
    answer = click.prompt(
        fmt.question(question),
        type=click.Choice(choices),
        prompt_suffix=" ",
        default=default,
        show_choices=False,
    )
    config[key] = answer



from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

from . import env
from . import fmt
from . import plugins


class BaseJobRunner:
    def __init__(self, root: str, config: Dict[str, Any]):
        self.root = root
        self.config = config

    def run_job_from_template(self, service: str, *path: str) -> None:
        command = self.render(*path)
        self.run_job(service, command)

    def run_cmd_from_template(self, service: str, *path: str) -> None:
        command = self.render(*path)
        self.run_cmd(service, command)

    def render(self, *path: str) -> str:
        rendered = env.render_file(self.config, *path).strip()
        if isinstance(rendered, bytes):
            raise TypeError("Cannot load job from binary file")
        return rendered

    def run_job(self, service: str, command: str) -> int:
        raise NotImplementedError

    def run_cmd(self, service: str, command: str) -> int:
        raise NotImplementedError

    def iter_plugin_hooks(
        self, hook: str
    ) -> Iterator[Tuple[str, Union[Dict[str, str], List[str]]]]:
        yield from plugins.iter_hooks(self.config, hook)


def initialise(runner: BaseJobRunner, limit_to: Optional[str] = None) -> None:
    fmt.echo_info("Initialising all services...")
    if limit_to is None or limit_to == "mysql":
        runner.run_job_from_template("mysql", "hooks", "mysql", "init")
    for plugin_name, hook in runner.iter_plugin_hooks("pre-init"):
        if limit_to is None or limit_to == plugin_name:
            for service in hook:
                fmt.echo_info(
                    "Plugin {}: running pre-init for service {}...".format(
                        plugin_name, service
                    )
                )
                runner.run_job_from_template(
                    service, plugin_name, "hooks", service, "pre-init"
                )
    for service in ["opencart"]:
        if limit_to is None or limit_to == service:
            fmt.echo_info("Initialising {}...".format(service))
            runner.run_cmd_from_template(service, "cmd", service, "init")
    for plugin_name, hook in runner.iter_plugin_hooks("init"):
        if limit_to is None or limit_to == plugin_name:
            for service in hook:
                fmt.echo_info(
                    "Plugin {}: running init for service {}...".format(
                        plugin_name, service
                    )
                )
                runner.run_job_from_template(
                    service, plugin_name, "hooks", service, "init"
                )
    fmt.echo_info("All services initialised.")



import os
from typing import Dict, Any

import click

from .. import config as peddler_config
from .. import env as peddler_env
from .. import fmt
from .. import utils
from . import compose
from .config import save as config_save_command
from .context import Context


def docker_compose(root: str, config: Dict[str, Any], *command: str) -> int:
    """
    Run docker-compose with local and production yml files.
    """
    args = []
    override_path = peddler_env.pathjoin(root, "local", "docker-compose.override.yml")
    if os.path.exists(override_path):
        args += ["-f", override_path]
    return utils.docker_compose(
        "-f",
        peddler_env.pathjoin(root, "local", "docker-compose.yml"),
        *args,
        "--project-name",
        config["LOCAL_PROJECT_NAME"],
        *command
    )


@click.group(help="Run OpenCart locally from scratch")
@click.pass_obj
def local(context: Context) -> None:
    context.docker_compose_func = docker_compose


@click.command(help="Configure and run OpenCart in docker-compose")
@click.option("-I", "--non-interactive", is_flag=True, help="Run non-interactively")
@click.option("-p", "--pullimages", is_flag=True, help="Update docker images")
@click.pass_context
def quickstart(context: click.Context, non_interactive: bool, pullimages: bool) -> None:
    click.echo(fmt.title("Interactive platform configuration"))
    context.invoke(
        config_save_command,
        interactive=(not non_interactive),
        set_vars=[],
        unset_vars=[],
    )
    click.echo(fmt.title("Stopping any existing platform"))
    context.invoke(compose.stop)
    if pullimages:
        click.echo(fmt.title("Docker image updates"))
        context.invoke(compose.dc_command, command="pull")
    click.echo(fmt.title("Starting the platform in detached mode"))
    context.invoke(compose.start, detach=True)
    click.echo(fmt.title("Database creation and migrations"))
    context.invoke(compose.init)

    config = peddler_config.load(context.obj.root)
    fmt.echo_info(
        """The OpenCart platform is now running in detached mode
Your OpenCart platform is ready and can be accessed at the following urls:

    {http}://{store_host}
    {http}://{store_host}admin
    """.format(
            http="https" if config["ENABLE_HTTPS"] else "http",
            store_host=config["STORE_HOST"],
        )
    )


local.add_command(quickstart)
compose.add_commands(local)



import io
import os

from setuptools import find_packages, setup

PWD = os.path.abspath(os.path.dirname(__file__))

def load_readme():
    with io.open(os.path.join(PWD, "README.rst"), "rt", encoding="utf8") as f:
        readme = f.read()
    # Replace img src for publication on pypi
    return readme.replace("./docs/img/", "https://github.com/alto9/peddler/raw/master/docs/img/")

def load_about():
    about = {}
    with io.open(
        os.path.join(PWD, "peddler", "__about__.py"), "rt", encoding="utf-8"
    ) as f:
        exec(f.read(), about)  # pylint: disable=exec-used
    return about

def load_requirements():
    with io.open(
        os.path.join(PWD, "requirements", "base.in"), "rt", encoding="utf-8"
    ) as f:
        return [line.strip() for line in f if is_requirement(line)]


def is_requirement(line):
    return not (line.strip() == "" or line.startswith("#"))

ABOUT = load_about()

setup(
    name="peddler",
    version=ABOUT["__version__"],
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    license="AGPLv3",
    install_requires=load_requirements(),
    entry_points={"console_scripts": ["peddler=peddler.commands.cli:main"]},
    author="alto9.com",
    author_email="contact@alto9.com",
    description="The Docker-based OpenCart distribution for control-freaks",
    long_description=load_readme(),
    long_description_content_type="text/x-rst",
)


import os
from typing import Any, Callable, Dict, List

import click
from mypy_extensions import VarArg

from .. import bindmounts
from .. import config as peddler_config
from .. import env as peddler_env
from ..exceptions import PeddlerError
from .. import fmt
from .. import jobs
from .. import serialize
from .. import utils
from .context import Context


class ComposeJobRunner(jobs.BaseJobRunner):
    def __init__(
        self,
        root: str,
        config: Dict[str, Any],
        docker_compose_func: Callable[[str, Dict[str, Any], VarArg(str)], int],
    ):
        super().__init__(root, config)
        self.docker_compose_func = docker_compose_func

    def run_job(self, service: str, command: str) -> int:
        """
        Run the "{{ service }}-job" service from local/docker-compose.jobs.yml with the
        specified command. For backward-compatibility reasons, if the corresponding
        service does not exist, run the service from good old regular
        docker-compose.yml.
        """
        jobs_path = peddler_env.pathjoin(self.root, "local", "docker-compose.jobs.yml")
        job_service_name = "{}-job".format(service)
        opts = [] if utils.is_a_tty() else ["-T"]
        if job_service_name in serialize.load(open(jobs_path).read())["services"]:
            return self.docker_compose_func(
                self.root,
                self.config,
                "-f",
                jobs_path,
                "run",
                *opts,
                "--rm",
                job_service_name,
                "sh",
                "-e",
                "-c",
                command,
            )
        fmt.echo_alert(
            (
                "The '{job_service_name}' service does not exist in {jobs_path}. "
                "This might be caused by an older plugin. Peddler switched to a job "
                "runner model for running one-time commands, such as database"
                " initialisation. For the record, this is the command that we are "
                "running:\n"
                "\n"
                "    {command}\n"
                "\n"
                "Old-style job running will be deprecated soon. Please inform "
                "your plugin maintainer!"
            ).format(
                job_service_name=job_service_name,
                jobs_path=jobs_path,
                command=command.replace("\n", "\n    "),
            )
        )
        return self.docker_compose_func(
            self.root,
            self.config,
            "run",
            *opts,
            "--rm",
            service,
            "sh",
            "-e",
            "-c",
            command,
        )

    def run_cmd(self, service: str, command: str) -> int:
        """
        Run the specified command on the "{{ service }}" service from local/docker-compose.yml.
        """
        compose_path = peddler_env.pathjoin(self.root, "local", "docker-compose.yml")

        opts = [] if utils.is_a_tty() else ["-T"]

        # if service in serialize.load(open(compose_path).read())["services"]:
        return self.docker_compose_func(
            self.root,
            self.config,
            "exec",
            *opts,
            service,
            "sh",
            "-e",
            "-c",
            command,
        )


@click.command(help="Run all or a selection of configured OpenCart services")
@click.option("-d", "--detach", is_flag=True, help="Start in daemon mode")
@click.argument("services", metavar="service", nargs=-1)
@click.pass_obj
def start(context: Context, detach: bool, services: List[str]) -> None:
    command = ["up", "--remove-orphans"]
    if detach:
        command.append("-d")

    config = peddler_config.load(context.root)
    context.docker_compose(context.root, config, *command, *services)


@click.command(help="Stop a running platform")
@click.argument("services", metavar="service", nargs=-1)
@click.pass_obj
def stop(context: Context, services: List[str]) -> None:
    config = peddler_config.load(context.root)
    context.docker_compose(context.root, config, "stop", *services)


@click.command(
    short_help="Reboot an existing platform",
    help="This is more than just a restart: with reboot, the platform is fully stopped before being restarted again",
)
@click.option("-d", "--detach", is_flag=True, help="Start in daemon mode")
@click.argument("services", metavar="service", nargs=-1)
@click.pass_context
def reboot(context: click.Context, detach: bool, services: List[str]) -> None:
    context.invoke(stop, services=services)
    context.invoke(start, detach=detach, services=services)


@click.command(
    short_help="Restart some components from a running platform.",
    help="""Specify 'store' to restart the store, or 'all' to
restart all services. Note that this performs a 'docker-compose restart', so new images
may not be taken into account. It is useful for reloading settings, for instance. To
fully stop the platform, use the 'reboot' command.""",
)
@click.argument("services", metavar="service", nargs=-1)
@click.pass_obj
def restart(context: Context, services: List[str]) -> None:
    config = peddler_config.load(context.root)
    command = ["restart"]
    if "all" in services:
        pass
    else:
        for service in services:
            if service == "store":
                if config["RUN_OPENCART"]:
                    command += ["store"]
            else:
                command.append(service)
    context.docker_compose(context.root, config, *command)


@click.command(help="Initialise all applications")
@click.option("-l", "--limit", help="Limit initialisation to this service or plugin")
@click.pass_obj
def init(context: Context, limit: str) -> None:
    config = peddler_config.load(context.root)
    runner = ComposeJobRunner(context.root, config, context.docker_compose)
    jobs.initialise(runner, limit_to=limit)


@click.command(
    help="Upload a design theme zip file. After upload, you will need to activate the theme in the OpenCart admin."
)
@click.argument("file_name")
@click.pass_obj
def upload_theme(context: Context, file_name: str) -> None:
    upload_path = peddler_env.pathjoin(context.root, "themes", file_name, "upload")
    if os.path.exists(upload_path):

        utils.docker("cp", upload_path, "opencart:/tmp/{}/".format(file_name))

        utils.docker(
            "exec",
            "-it",
            "opencart",
            "cp",
            "-r",
            "/tmp/{}/.".format(file_name),
            "/var/www/html/",
        )
    else:
        fmt.echo_alert("No theme upload folder was found at {}.".format(upload_path))


@click.command(
    short_help="Run a command in a new container",
    help=(
        "Run a command in a new container. This is a wrapper around `docker-compose run`. Any option or argument passed"
        " to this command will be forwarded to docker-compose. Thus, you may use `-v` or `-p` to mount volumes and"
        " expose ports."
    ),
    context_settings={"ignore_unknown_options": True},
)
@click.argument("args", nargs=-1, required=True)
@click.pass_context
def run(context: click.Context, args: List[str]) -> None:
    extra_args = ["--rm"]
    if not utils.is_a_tty():
        extra_args.append("-T")
    context.invoke(dc_command, command="run", args=[*extra_args, *args])


@click.command(
    name="bindmount",
    help="Copy the contents of a container directory to a ready-to-bind-mount host directory",
)
@click.argument(
    "service",
)
@click.argument("path")
@click.pass_obj
def bindmount_command(context: Context, service: str, path: str) -> None:
    config = peddler_config.load(context.root)
    host_path = bindmounts.create(
        context.root, config, context.docker_compose, service, path
    )
    fmt.echo_info(
        "Bind-mount volume created at {}. You can now use it in all `local` and `dev` commands with the `--volume={}` option.".format(
            host_path, path
        )
    )


@click.command(
    short_help="Run a command in a running container",
    help=(
        "Run a command in a running container. This is a wrapper around `docker-compose exec`. Any option or argument"
        " passed to this command will be forwarded to docker-compose. Thus, you may use `-e` to manually define"
        " environment variables."
    ),
    context_settings={"ignore_unknown_options": True},
    name="exec",
)
@click.argument("args", nargs=-1, required=True)
@click.pass_context
def execute(context: click.Context, args: List[str]) -> None:
    context.invoke(dc_command, command="exec", args=args)


@click.command(
    short_help="View output from containers",
    help="View output from containers. This is a wrapper around `docker-compose logs`.",
)
@click.option("-f", "--follow", is_flag=True, help="Follow log output")
@click.option("--tail", type=int, help="Number of lines to show from each container")
@click.argument("service", nargs=-1)
@click.pass_context
def logs(context: click.Context, follow: bool, tail: bool, service: str) -> None:
    args = []
    if follow:
        args.append("--follow")
    if tail is not None:
        args += ["--tail", str(tail)]
    args += service
    context.invoke(dc_command, command="logs", args=args)


@click.command(
    short_help="Direct interface to docker-compose.",
    help=(
        "Direct interface to docker-compose. This is a wrapper around `docker-compose`. Most commands, options and"
        " arguments passed to this command will be forwarded as-is to docker-compose."
    ),
    context_settings={"ignore_unknown_options": True},
    name="dc",
)
@click.argument("command")
@click.argument("args", nargs=-1, required=True)
@click.pass_obj
def dc_command(context: Context, command: str, args: List[str]) -> None:
    config = peddler_config.load(context.root)
    volumes, non_volume_args = bindmounts.parse_volumes(args)
    volume_args = []
    for volume_arg in volumes:
        if ":" not in volume_arg:
            # This is a bind-mounted volume from the "volumes/" folder.
            host_bind_path = bindmounts.get_path(context.root, volume_arg)
            if not os.path.exists(host_bind_path):
                raise PeddlerError(
                    (
                        "Bind-mount volume directory {} does not exist. It must first be created"
                        " with the '{}' command."
                    ).format(host_bind_path, bindmount_command.name)
                )
            volume_arg = "{}:{}".format(host_bind_path, volume_arg)
        volume_args += ["--volume", volume_arg]
    context.docker_compose(
        context.root, config, command, *volume_args, *non_volume_args
    )


def add_commands(command_group: click.Group) -> None:
    command_group.add_command(start)
    command_group.add_command(stop)
    command_group.add_command(restart)
    command_group.add_command(reboot)
    command_group.add_command(init)
    command_group.add_command(upload_theme)
    command_group.add_command(dc_command)
    command_group.add_command(run)
    command_group.add_command(bindmount_command)
    command_group.add_command(execute)
    command_group.add_command(logs)
